#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
    Outbound Message Factory should be implemented to serialize messages
    in a format that is expected to be received from the platform.
"""


class OutboundMessageFactory:
    """
        Outbound Message Factory Interface
    """

    def make_from_sensor_reading(self, reading):
        """
        Serializes a sensor reading to be sent to the platform

        :param reading: The reading to be serialized
        :type reading: SensorReading
        :returns: message
        :rtype: OutboundMessage
        """
        pass

    def make_from_alarm(self, alarm):
        """
        Serializes an alarm event to be sent to the platform

        :param alarm: The alarm to be serialized
        :type alarm: Alarm
        :returns: message
        :rtype: OutboundMessage
        """
        pass

    def make_from_actuator_status(self, actuator):
        """
        Serializes an actuator status to be sent to the platform

        :param actuator: The actuator status to be serialized
        :type actuator: ActuatorStatus
        :returns: message
        :rtype: OutboundMessage
        """
        pass

    def make_from_firmware_status(self, firmware_status):
        """
        Reports the current status of the firmware update process

        :param firmware_status: The current status of the firmware update process
        :type firmware_status: FirmwareStatus
        :returns: message
        :rtype: OutboundMessage
        """
        pass

    def make_from_chunk_request(self, file_name, chunk_index, chunk_size):
        """
        Requests a chunk of the firmware file from the platform

        :param file_name: The name of the file that contains the requested chunk
        :type file_name: string
        :param chunk_index: The index of the requested chunk
        :type chunk_index: int
        :param chunk_size: The size of the requested chunk
        :type chunk_size: int
        :returns: message
        :rtype: OutboundMessage
        """
        pass

    def make_from_firmware_version(self, version):
        """
        Reports the current firmware version to the platform

        :param version: Current firmware version
        :type version: str
        :returns: message
        :rtype: OutboundMessage
        """
        pass

    def make_from_keep_alive_message(self):
        """
        Sends a ping to the platform

        :returns: message
        :rtype: OutboundMessage
        """
        pass

    def make_from_configuration(self, configuration):
        """
        Serializes the device's configuration to be sent to the platform

        :param configuration: The device's current configuration
        :type configuration: dict
        :returns: message
        :rtype: OutboundMessage
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
    Sensor as defined in the device manifest
"""


class SensorReading:
    """
        Sensor Reading Class
    """

    def __init__(self, reference, value, timestamp=None):
        """
        :param reference: The reference of the sensor
        :type reference: str
        :param value: The value of the reading
        :type value: int, float, str
        :param timestamp: (optional) Unix timestamp - if not provided, platform will assign one upon reception
        :type timestamp: int
        """
        self.reference = reference
        self.value = value
        self.timestamp = timestamp



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Configuration Provider module.

Contains ConfigurationProvider "interface".
"""


class ConfigurationProvider:
    """
    Must be implemented to read the device's configuration values.

    The configuration values are then sent to WolkAbout IoT Platform.
    """

    def get_configuration(self):
        """
        Read device configuration and return it as a dictionary.

        With device configuration reference as key,
        and device configuration value as value.

        Must be implemented as non blocking.
        Must be implemented as thread safe.

        :returns: configuration
        :rtype: dict
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
    Inbound Messages that get received from the platform
"""


class InboundMessage:
    """
        Inbound Messages class
    """

    def __init__(self, channel, payload):
        """
        :param channel: Defines the channel where the message was published to
        :type channel: str
        :param payload: Defines the payload of the message that was published to the channel
        :type payload: str
        """
        self.channel = channel
        self.payload = payload



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from wolk.wolkcore import SensorReading
from wolk.wolkcore import Alarm
from wolk.wolkcore import ActuatorCommandType
from wolk.wolkcore import ActuatorStatus
from wolk.wolkcore import ConfigurationCommandType
from wolk.wolkcore import FirmwareCommandType
from wolk.wolkcore import FirmwareStatus
from wolk.wolkcore import FirmwareStatusType
from wolk.wolkcore import FirmwareErrorType

"""
    WolkCore Class
"""


class WolkCore:
    """
        WolkCore Class
    """

    def __init__(
        self,
        outbound_message_factory,
        outbound_message_queue,
        connectivity_service,
        actuation_handler,
        actuator_status_provider,
        inbound_message_deserializer,
        configuration_handler,
        configuration_provider,
        keep_alive_service=None,
        firmware_update=None,
    ):
        """
        :param outbound_message_factory: Implementation of the OutboundMessageFactory interface
        :type outbound_message_factory: OutboundMessageFactory
        :param outbound_message_queue: Implementation of the OutboundMessageQueue interface
        :type outbound_message_queue: OutboundMessageQueue
        :param connectivity_service: Implementation of the ConnectivityService interface
        :type connectivity_service: ConnectivityService
        :param actuation_handler: Implementation of the ActuationHandler interface
        :type actuation_handler: ActuationHandler
        :param actuator_status_provider: Implementation of the ActuatorStatusProvider interface
        :type actuator_status_provider: ActuatorStatusProvider
        :param inbound_message_deserializer: Implementation of the InboundMessageDeserializer interface
        :type inbound_message_deserializer: InboundMessageDeserializer
        :param configuration_handler: Implementation of the ConfigurationHandler interface
        :type configuration_handler: ConfigurationHandler
        :param configuration_provider: Implementation of the ConfigurationProvider interface
        :type configuration_provider: ConfigurationProvider
        :param keep_alive_service: Implementation of the KeepAliveService interface
        :type keep_alive_service: KeepAliveService
        :param firmware_update: Implementation of the FirmwareUpdate interface
        :type firmware_update: FirmwareUpdate
        """

        self.outbound_message_factory = outbound_message_factory
        self.outbound_message_queue = outbound_message_queue
        self.connectivity_service = connectivity_service
        self.inbound_message_deserializer = inbound_message_deserializer
        self.actuation_handler = actuation_handler
        self.actuator_status_provider = actuator_status_provider
        self.configuration_handler = configuration_handler
        self.configuration_provider = configuration_provider
        self.connectivity_service.set_inbound_message_listener(self._on_inbound_message)

        self.keep_alive_service = None

        if keep_alive_service:
            self.keep_alive_service = keep_alive_service

        self.firmware_update = firmware_update

        if self.firmware_update:

            self.firmware_update.set_on_file_packet_request_callback(
                self._on_packet_request
            )
            self.firmware_update.set_on_status_callback(self._on_status)

    def connect(self):
        """
        Connects to the platform and starts the keep alive service if it is enabled
        """
        self.connectivity_service.connect()
        if self.keep_alive_service:
            self.keep_alive_service.start()

    def disconnect(self):
        """
        Disconnects from the platform and stops the keep alive service if it is enabled
        """
        self.connectivity_service.disconnect()
        if self.keep_alive_service:
            self.keep_alive_service.stop()

    def add_sensor_reading(self, reference, value, timestamp=None):
        """
        Publish a sensor reading to the platform
        :param reference: The reference of the sensor
        :type reference: str
        :param value: The value of the sensor reading
        :type value: int, float, str
        :param timestamp: (optional) Unix timestamp - if not provided, platform will assign one upon reception
        :type timestamp: int
        """
        reading = SensorReading(reference, value, timestamp)
        outbound_message = self.outbound_message_factory.make_from_sensor_reading(
            reading
        )
        self.outbound_message_queue.put(outbound_message)

    def add_alarm(self, reference, message, timestamp=None):
        """
        Publish an alarm to the platform
        :param reference: The reference of the alarm
        :type reference: str
        :param message: Description of the event that occurred
        :type message: str
        :param timestamp: (optional) Unix timestamp - if not provided, platform will assign one upon reception
        :type timestamp: int
        """
        alarm = Alarm(reference, message, timestamp)
        outbound_message = self.outbound_message_factory.make_from_alarm(alarm)
        self.outbound_message_queue.put(outbound_message)

    def publish(self):
        """
        Publishes all currently stored messages to the platform
        """
        while True:
            outbound_message = self.outbound_message_queue.peek()

            if outbound_message is None:

                break

            if self.connectivity_service.publish(outbound_message) is True:

                self.outbound_message_queue.get()

            else:

                break

    def publish_actuator_status(self, reference):
        """
        Publish the current actuator status to the platform
        :param reference: The reference of the actuator
        :type reference: str
        """
        state, value = self.actuator_status_provider.get_actuator_status(reference)
        actuator_status = ActuatorStatus(reference, state, value)
        outbound_message = self.outbound_message_factory.make_from_actuator_status(
            actuator_status
        )

        if not self.connectivity_service.publish(outbound_message):
            self.outbound_message_queue.put(outbound_message)

    def publish_configuration(self):
        """
        Publish the current device configuration to the platform
        """
        configuration = self.configuration_provider.get_configuration()
        outbound_message = self.outbound_message_factory.make_from_configuration(
            configuration
        )

        if not self.connectivity_service.publish(outbound_message):
            self.outbound_message_queue.put(outbound_message)

    def _on_inbound_message(self, message):
        """
        Callback function to handle inbound messages
        .. note:: Pass this function to the implementation of ConnectivityService
        :param message:  The message received from the platform
        :type message: InboundMessage
        """
        if message.channel.startswith("actuators/commands/"):

            if not self.actuation_handler or not self.actuator_status_provider:
                return

            actuation = self.inbound_message_deserializer.deserialize_actuator_command(
                message
            )

            if actuation.command == ActuatorCommandType.ACTUATOR_COMMAND_TYPE_SET:

                self.actuation_handler.handle_actuation(
                    actuation.reference, actuation.value
                )

                self.publish_actuator_status(actuation.reference)

            elif actuation.command == ActuatorCommandType.ACTUATOR_COMMAND_TYPE_STATUS:

                self.publish_actuator_status(actuation.reference)

        elif message.channel.startswith("service/commands/firmware/"):

            if not self.firmware_update:
                # Firmware update disabled
                firmware_status = FirmwareStatus(
                    FirmwareStatusType.FIRMWARE_STATUS_ERROR,
                    FirmwareErrorType.FIRMWARE_ERROR_FILE_UPLOAD_DISABLED,
                )
                outbound_message = self.outbound_message_factory.make_from_firmware_status(
                    firmware_status
                )
                if not self.connectivity_service.publish(outbound_message):
                    self.outbound_message_queue.put(outbound_message)
                return

            firmware_command = self.inbound_message_deserializer.deserialize_firmware_command(
                message
            )

            if (
                firmware_command.command
                == FirmwareCommandType.FIRMWARE_COMMAND_TYPE_FILE_UPLOAD
            ):

                self.firmware_update.handle_file_upload(firmware_command)

            elif (
                firmware_command.command
                == FirmwareCommandType.FIRMWARE_COMMAND_TYPE_URL_DOWNLOAD
            ):

                self.firmware_update.handle_url_download(firmware_command)

            elif (
                firmware_command.command
                == FirmwareCommandType.FIRMWARE_COMMAND_TYPE_INSTALL
            ):

                self.firmware_update.handle_install()

            elif (
                firmware_command.command
                == FirmwareCommandType.FIRMWARE_COMMAND_TYPE_ABORT
            ):

                self.firmware_update.handle_abort()

            elif (
                firmware_command.command
                == FirmwareCommandType.FIRMWARE_COMMAND_TYPE_UNKNOWN
            ):
                pass

        elif message.channel.startswith("service/binary/"):

            if not self.firmware_update:
                # Firmware update disabled
                firmware_status = FirmwareStatus(
                    FirmwareStatusType.FIRMWARE_STATUS_ERROR,
                    FirmwareErrorType.FIRMWARE_ERROR_FILE_UPLOAD_DISABLED,
                )
                outbound_message = self.outbound_message_factory.make_from_firmware_status(
                    firmware_status
                )
                if not self.connectivity_service.publish(outbound_message):
                    self.outbound_message_queue.put(outbound_message)
                return

            packet = self.inbound_message_deserializer.deserialize_firmware_chunk(
                message
            )
            self.firmware_update.handle_packet(packet)

        elif message.channel.startswith("configurations/commands/"):

            if not self.configuration_provider or not self.configuration_handler:
                return

            configuration = self.inbound_message_deserializer.deserialize_configuration_command(
                message
            )

            if (
                configuration.command
                == ConfigurationCommandType.CONFIGURATION_COMMAND_TYPE_SET
            ):
                self.configuration_handler.handle_configuration(configuration.values)
                self.publish_configuration()

            elif (
                configuration.command
                == ConfigurationCommandType.CONFIGURATION_COMMAND_TYPE_CURRENT
            ):
                self.publish_configuration()

    def _on_packet_request(self, file_name, chunk_index, chunk_size):
        """
        """
        message = self.outbound_message_factory.make_from_chunk_request(
            file_name, chunk_index, chunk_size
        )
        if not self.connectivity_service.publish(message):
            self.outbound_message_queue.put(message)

    def _on_status(self, status):
        """
        """
        message = self.outbound_message_factory.make_from_firmware_status(status)
        if not self.connectivity_service.publish(message):
            self.outbound_message_queue.put(message)



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
    Contains the status of the firmware update process.
"""


class FirmwareStatus:
    """
    Firmware Status class
    """

    def __init__(self, status, error=None):
        """
        :param status: The status of the firmware update process
        :type status: FirmwareStatusType
        :param error: The type of error that occured
        :type error: FirmwareErrorType
        """
        self.status = status
        self.error = error



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
File System Firmware Handler module.

Contains FileSystemFirmwareHandler class used for storing firmware file.
"""
import os
import shutil
import tempfile
from threading import Timer
from urllib.parse import urlparse

from persistent_queue import PersistentQueue

from wolk import LoggerFactory
from wolk.wolkcore import FirmwareHandler


class FileSystemFirmwareHandler(FirmwareHandler):
    """
    Implementation of FirmwareHandler interface that stores the file on disk.

    :ivar chunk_size: desired chunk size in bytes
    :vartype chunk_size: int
    :ivar download_location: where to store the completed firmware file
    :vartype download_location: str
    :ivar file_name: name of firmware file
    :vartype file_name: str
    :ivar file_size: size of firmware file in bytes
    :vartype file_size: int
    :ivar file_url: URL where there firmware file is located
    :vartype file_url: str
    :ivar firmware_installer: implementation of FirmwareInstaller
    :vartype firmware_installer: FirmwareInstaller
    :ivar firmware_url_download_handler: URL downloader
    :vartype firmware_url_download_handler: FirmwareURLDownloadHandler
    :ivar logger: Logger instance issued from the LoggerFactory class
    :vartype logger: logger
    :ivar max_file_size: maximum file size supported by device in bytes
    :vartype max_file_size: int
    :ivar report_result_callback: Callback for reporting URL download result
    :vartype report_result_callback: method
    :ivar temp_file: handle of temp file used to store incomplete firmware file
    :vartype temp_file: file object
    :ivar version: current version of the device firmware
    :vartype version: str
    :ivar version_persister: means of storing current version on disk
    :vartype version_persister: PersistentQueue
    """

    def __init__(
        self,
        version,
        chunk_size,
        max_file_size,
        firmware_installer,
        download_location,
        firmware_url_download_handler=None,
    ):
        """
        Handle file manipulation on disk.

        :param version: current version of the device firmware
        :type version: str
        :param chunk_size: desired chunk size in bytes
        :type chunk_size: int
        :param max_file_size: maximum file size supported by device in bytes
        :type max_file_size: int
        :param firmware_installer: implementation of FirmwareInstaller
        :type firmware_installer: FirmwareInstaller
        :param download_location: where to store the completed firmware file
        :type download_location: str
        :param firmware_url_download_handler: URL downloader
        :type firmware_url_download_handler: FirmwareURLDownloadHandler or None
        """
        self.logger = LoggerFactory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(
            "Init:  Version: %s ; Chunk size: %s ; Max file size: %s ; "
            "Firmware installer: %s ; Download location: '%s'  "
            "Firmware URL download handler: %s",
            version,
            chunk_size,
            max_file_size,
            firmware_installer,
            download_location,
            firmware_url_download_handler,
        )
        self.version = version
        self.chunk_size = chunk_size
        self.max_file_size = max_file_size
        self.download_location = download_location
        self.firmware_installer = firmware_installer
        self.firmware_url_download_handler = firmware_url_download_handler
        self.temp_file = None
        self.file_name = None
        self.file_size = None
        self.file_url = None
        self.report_result_callback = None
        self.version_persister = PersistentQueue("persisted_version")

    def update_start(self, file_name, file_size):
        """
        Start receiving the firmware file.

        :param file_name: name of the firmware file
        :type file_name: str
        :param file_size: size of the firmware file in bytes
        :type file_size: int

        :returns: result
        :rtype: bool
        """
        self.logger.debug(
            "update_start called - File name: %s ; File size: %s",
            file_name,
            file_size,
        )
        if file_size > self.max_file_size:
            self.logger.debug(
                "update_start - File size too big, returning False"
            )
            return False

        self.temp_file = tempfile.NamedTemporaryFile(mode="a+b", delete=False)
        self.file_name = file_name
        self.file_size = file_size
        self.logger.debug(
            "update_start - Temporary file created, returning True"
        )
        return True

    def update_finalize(self):
        """
        Finalize firmware installation process.

        Copies the content of the temporary file to
        the desired download location.
        Calls the provided firmware installer's install_firmware function
        """
        self.logger.debug("update_finalize called")

        if not os.path.exists(os.path.abspath(self.download_location)):
            os.makedirs(os.path.abspath(self.download_location))

        firmware_file_path = os.path.join(
            os.path.abspath(self.download_location), self.file_name
        )

        shutil.copy2(os.path.realpath(self.temp_file.name), firmware_file_path)

        if self.temp_file:
            self.temp_file.close()

        self.logger.info(
            "Firmware file copied to download location from "
            "temporary file, calling firmware_installer.install_firmware "
            "with path: %s",
            firmware_file_path,
        )
        self.firmware_installer.install_firmware(firmware_file_path)

    def update_abort(self):
        """Abort the firmware update process."""
        self.logger.debug("update_abort called")

        if self.temp_file:
            self.temp_file.close()

        self.temp_file = None
        self.file_name = None
        self.file_size = None
        self.file_url = None

    def write_chunk(self, chunk):
        """
        Write a chunk of the firmware file to the temporary file.

        :param chunk: A piece of the firmware file
        :type chunk: Bytes

        :returns: result
        :rtype: bool
        """
        self.logger.debug("write_chunk called - Chunk size: %s", len(chunk))

        self.temp_file.write(chunk)
        self.temp_file.flush()
        os.fsync(self.temp_file)
        self.logger.debug("write_chunk - Chunk written, returning True")
        return True

    def read_chunk(self, index):
        """
        Read a chunk of the temporary firmware file.

        :param index: offset from the beginning of the file
        :type index: int

        :returns: chunk
        :rtype: bytes
        """
        self.logger.debug("read_chunk called - Index: %s", index)

        self.temp_file.seek(index * self.chunk_size)
        chunk = self.temp_file.read(self.chunk_size)
        self.logger.debug("read_chunk - Chunk size: %s", len(chunk))
        return chunk

    def persist_version(self, version):
        """
        Place the current firmware version into persistent storage.

        Later to be used to determine the result
        of the firmware update process

        :param version: The current firmware version
        :type version: str

        :returns: result
        :rtype: bool
        """
        self.logger.debug("persist_version called - Version: %s", version)

        self.version_persister.clear()
        self.version_persister.flush()
        self.version_persister.push(version)
        self.logger.debug(
            "persist_version - Persisted version, returning True"
        )
        return True

    def unpersist_version(self):
        """
        Remove the firmware version from persistent storage.

        :returns: version
        :rtype: str or None
        """
        self.logger.debug("unpersist_version called")

        if not self.version_persister.peek():
            self.version_persister.clear()
            self.logger.debug(
                "unpersist_version - No persisted version, returning None"
            )
            return None
        else:
            version = self.version_persister.pop()
            self.version_persister.flush()
            self.logger.debug(
                "unpersist_version - Persisted version found, returning %s",
                version,
            )
            return version

    def set_url_download_result_callback(self, callback):
        """
        Set callback for reporting URL download result.

        :param callback: method to call
        :type callback: method
        """
        self.logger.debug(
            "set_url_download_result_callback called - Callback: %s", callback
        )
        self.report_result_callback = callback

    def update_start_url_download(self, file_url):
        """
        Start firmware file URL download process.

        Calls download function from firmware_url_download_handler.
        Returns the validity of the URL and calls download method if valid.

        :param file_url: The URL that contains the firmware file
        :type file_url: str

        :returns: result
        :rtype: bool
        """
        self.logger.debug(
            "update_start_url_download called - File URL: %s", file_url
        )

        if not self.firmware_url_download_handler:
            self.logger.debug(
                "update_start_url_download - No firmware_url_download_handler,"
                "returning False"
            )
            return False

        parsed_url = urlparse(file_url)

        if bool(parsed_url.scheme):

            self.file_url = file_url
            self.file_name = self.file_url.split("/")[-1]
            t = Timer(
                2.0,
                self.firmware_url_download_handler.download,
                [self.file_url, self.file_name, self.report_result_callback],
            )
            t.start()
            self.logger.debug(
                "update_start_url_download - Valid URL, calling "
                "firmware_url_download_handler.download and returning True"
            )
            return True

        else:

            self.logger.debug(
                "update_start_url_download - Invalid URL, returning False"
            )
            return False



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


"""
    Deserialize the messages received from the platform
"""


class InboundMessageDeserializer:
    """
        Inbound Message Deserializer Interface
    """

    def deserialize_actuator_command(self, message):
        """
        Deserialize the message into an ActuatorCommand

        :param message: The message received
        :type message: InboundMessage
        :returns: actuation
        :rtype: ActuatorCommand
        """
        pass

    def deserialize_firmware_command(self, message):
        """
        Deserializes the message into a FirmwareCommand

        :param message: The message received
        :type message: InboundMessage
        :returns: firmware_command
        :rtype: FirmwareCommand
        """
        pass

    def deserialize_firmware_chunk(self, message):
        """
        Deserializes the message into a tuple of bytes

        :param message: The message received
        :type message: InboundMessage
        :returns: (previous_hash, chunk, chunk_hash)
        :rtype: (bytes, bytes, bytes)
        """
        pass

    def deserialize_configuration_command(self, message):
        """
        Deserializes the message into a ConfigurationCommand

        :param message: The message received
        :type message: InboundMessage
        :returns: configuration
        :rtype: ConfigurationCommand
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Actuation Handler module.

Contains ActuationHandler "interface".
"""


class ActuationHandler:
    """
    Must be implemented in order to execute actuation commands.

    Commands are issued from WolkAbout IoT Platform.
    """

    def handle_actuation(self, reference, value):
        """
        Set the actuator, identified by reference, to the value specified.

        :param reference: reference of the actuator
        :type reference: str
        :param value: value to set to
        :type value: int or float or str
        """
        pass



# coding=utf-8

from .ActuatorCommand import ActuatorCommand
from .ActuationHandler import ActuationHandler
from .ActuatorCommandType import *
from .ActuatorState import *
from .ActuatorStatus import ActuatorStatus
from .ActuatorStatusProvider import ActuatorStatusProvider
from .Alarm import Alarm
from .ConfigurationCommand import ConfigurationCommand
from .ConfigurationCommandType import *
from .ConfigurationHandler import ConfigurationHandler
from .ConfigurationProvider import ConfigurationProvider
from .ConnectivityService import ConnectivityService
from .Device import Device
from .FileTransferPacket import FileTransferPacket
from .FirmwareCommand import FirmwareCommand
from .FirmwareCommandType import *
from .FirmwareErrorType import *
from .FirmwareHandler import FirmwareHandler
from .FirmwareStatus import FirmwareStatus
from .FirmwareStatusType import *
from .FirmwareUpdate import FirmwareUpdate
from .FirmwareUpdateStateType import *
from .InboundMessage import InboundMessage
from .InboundMessageDeserializer import InboundMessageDeserializer
from .KeepAliveService import KeepAliveService
from .OutboundMessage import OutboundMessage
from .OutboundMessageFactory import OutboundMessageFactory
from .OutboundMessageQueue import OutboundMessageQueue
from .SensorReading import SensorReading
from .Version import *
from .WolkCore import WolkCore

__all__ = [
    "ActuationHandler",
    "ActuatorCommand",
    "ActuatorCommandType",
    "ActuatorStatus",
    "ActuatorStatusProvider",
    "Alarm",
    "ConfigurationCommand",
    "ConfigurationCommandType",
    "ConfigurationHandler",
    "ConfigurationProvider",
    "ConnectivityService",
    "Device",
    "FileTransferPacket",
    "FirmwareCommand",
    "FirmwareCommandType",
    "FirmwareErrorType",
    "FirmwareHandler",
    "FirmwareStatus",
    "FirmwareStatusType",
    "FirmwareUpdate",
    "FirmwareUpdateStateType",
    "InboundMessage",
    "InboundMessageDeserializer",
    "KeepAliveService",
    "OutboundMessage",
    "OutboundMessageFactory",
    "OutboundMessageQueue",
    "SensorReading",
    "WolkCore",
]



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
WolkAbout IoT Platform Library.

Contains WolkConnect class that represents the top level API.
"""
import os

from wolk.OSOutboundMessageQueue import OSOutboundMessageQueue
from wolk.OSMQTTConnectivityService import OSMQTTConnectivityService
from wolk.OSOutboundMessageFactory import OSOutboundMessageFactory
from wolk.OSInboundMessageDeserializer import OSInboundMessageDeserializer
from wolk.OSFirmwareUpdate import OSFirmwareUpdate
from wolk.OSKeepAliveService import OSKeepAliveService
from wolk.LoggerFactory import logger_factory

from wolk.wolkcore import WolkCore


class WolkConnect:
    """
    Send sensor readings, actuator statuses and data to WolkAbout IoT platform.

    :ivar connectivity_service: means of sending/receiving data
    :vartype connectivity_service: OSMQTTConnectivityService
    :ivar deserializer: deserializer of inbound messages
    :vartype deserializer: OSInboundMessageDeserializer
    :ivar device: Contains device key and password, and actuator references
    :vartype device: Device
    :ivar firmware_update: Firmware update handler
    :vartype firmware_update: OSFirmwareUpdate
    :ivar keep_alive_service: Keep device connected when not sending data
    :vartype keep_alive_service: OSKeepAliveService
    :ivar logger: Logger instance issued by the LoggerFactory class
    :vartype logger: logger
    :ivar outbound_message_factory: Create messages to send
    :vartype outbound_message_factory: OSOutboundMessageFactory
    :ivar outbound_message_queue: Store data before sending
    :vartype outbound_message_queue: OSOutboundMessageQueue
    """

    def __init__(
        self,
        device,
        actuation_handler=None,
        actuator_status_provider=None,
        configuration_handler=None,
        configuration_provider=None,
        outbound_message_queue=OSOutboundMessageQueue(),
        keep_alive_enabled=True,
        firmware_handler=None,
        host=None,
        port=None,
        ca_cert=None,
    ):
        """
        Provide communication with WolkAbout IoT Platform.

        :param device: Contains key and password, and actuator references
        :type device: Device
        :param actuation_handler: Handle actuation commands
        :type actuation_handler: ActuationHandler, optional
        :param actuator_status_provider: Read actuator status
        :type actuator_status_provider: ActuatorStatusProvider, optional
        :param outbound_message_queue: Store messages before sending
        :type outbound_message_queue: OutboundMessageQueue, optional
        :param keep_alive_enabled: Enable or disable the keep alive service
        :type keep_alive_enabled: bool, default True
        :param firmware_handler: Firmware update support
        :type firmware_handler: FirmwareHandler, optional
        :param host: The host name or IP address of the remote broker
        :type host: str, optional
        :param port: The network port of the server host to connect to
        :type port: int, optional
        :param ca_cert: String path to Certificate Authority certificate file
        :type ca_cert: str, optional
        """
        self.logger = logger_factory.get_logger(str(self.__class__.__name__))
        self.logger.debug(
            "Init:  Device: %s ; Actuation handler: %s ; "
            "Actuator status provider: %s ; "
            "Configuration handler: %s ; Configuration provider: %s ; "
            "Outbound message queue: %s ; Keep alive enabled: %s ;"
            " Firmware handler: %s"
            "Host: %s ; Port: %s ; ca_cert: %s",
            device,
            actuation_handler,
            actuator_status_provider,
            configuration_handler,
            configuration_provider,
            outbound_message_queue,
            keep_alive_enabled,
            firmware_handler,
            host,
            port,
            ca_cert,
        )
        self.device = device
        self.outbound_message_factory = OSOutboundMessageFactory(device.key)
        self.outbound_message_queue = outbound_message_queue
        if host and port:
            if ca_cert:
                self.connectivity_service = OSMQTTConnectivityService(
                    device, host=host, port=int(port), ca_cert=ca_cert
                )
            else:
                self.connectivity_service = OSMQTTConnectivityService(
                    device, host=host, port=int(port)
                )

        else:
            wolk_ca_cert = os.path.join(os.path.dirname(__file__), "ca.crt")
            self.connectivity_service = OSMQTTConnectivityService(
                device, ca_cert=wolk_ca_cert
            )
        self.deserializer = OSInboundMessageDeserializer()
        self.firmware_update = OSFirmwareUpdate(firmware_handler)

        self.keep_alive_service = None
        if keep_alive_enabled:
            keep_alive_interval_seconds = 600
            self.logger.debug(
                "Keep alive enabled, interval: %s", keep_alive_interval_seconds
            )
            self.keep_alive_service = OSKeepAliveService(
                self.connectivity_service,
                self.outbound_message_factory,
                keep_alive_interval_seconds,
            )

        if device.actuator_references and (
            actuation_handler is None or actuator_status_provider is None
        ):
            raise RuntimeError(
                "Both ActuatorStatusProvider and ActuationHandler "
                "must be provided for device with actuators."
            )

        self._wolk = WolkCore(
            self.outbound_message_factory,
            self.outbound_message_queue,
            self.connectivity_service,
            actuation_handler,
            actuator_status_provider,
            self.deserializer,
            configuration_handler,
            configuration_provider,
            self.keep_alive_service,
            self.firmware_update,
        )

        if firmware_handler:

            message = self.outbound_message_factory.make_from_firmware_version(
                firmware_handler.version
            )
            self.outbound_message_queue.put(message)

        self.firmware_update.report_result()

    def connect(self):
        """
        Connect the device to the WolkAbout IoT Platform.

        :raises e: Exception from MQTT client
        """
        try:

            self.logger.debug("connect started")
            self._wolk.connect()
            self.logger.debug("connect ended")

        except Exception as e:

            raise e

    def disconnect(self):
        """Disconnect the device from the WolkAbout IoT Platform."""
        self.logger.debug("disconnect started")
        self._wolk.disconnect()
        self.logger.debug("disconnect ended")

    def add_sensor_reading(self, reference, value, timestamp=None):
        """
        Publish a sensor reading to WolkAbout IoT Platform.

        :param reference: The reference of the sensor
        :type reference: str
        :param value: The value of the sensor reading
        :type value: int, float
        :param timestamp: Unix timestamp. If not provided, platform will assign
        :type timestamp: int, optional
        """
        self.logger.debug("add_sensor_reading started")
        self._wolk.add_sensor_reading(reference, value, timestamp)
        self.logger.debug("add_sensor_reading ended")

    def add_alarm(self, reference, active, timestamp=None):
        """
        Publish an alarm to the WolkAbout IoT Platform.

        :param reference: Reference of the alarm
        :type reference: str
        :param active: Current state of the alarm
        :type active: bool
        :param timestamp: Unix timestamp. If not provided, platform will assign
        :type timestamp: int
        """
        self.logger.debug("add_alarm started")
        self._wolk.add_alarm(reference, active, timestamp)
        self.logger.debug("add_alarm ended")

    def publish(self):
        """Publish all currently stored messages to WolkAbout IoT Platform."""
        self.logger.debug("publish started")

        self._wolk.publish()
        self.logger.debug("publish ended")

    def publish_actuator_status(self, reference):
        """
        Publish the current actuator status to the WolkAbout IoT Platform.

        :param reference: reference of the actuator
        :type reference: str
        """
        self.logger.debug("publish_actuator_status started")
        self._wolk.publish_actuator_status(reference)
        self.logger.debug("publish_actuator_status ended")

    def publish_configuration(self):
        """Publish the current device configuration to the platform."""
        self.logger.debug("publish_configuration started")
        self._wolk.publish_configuration()
        self.logger.debug("publish_configuration ended")

    def _on_inbound_message(self, message):
        """
        Handle inbound messages.

        :param message: message received from the platform
        :type message: InboundMessage
        """
        self.logger.debug("_on_inbound_message started")
        self._wolk._on_inbound_message(message)
        self.logger.debug("_on_inbound_message ended")

    def _on_packet_request(self, file_name, chunk_index, chunk_size):
        """
        Handle firmware packet requests.

        :param file_name: The name of the file to which the chunk belongs
        :type file_name: str
        :param chunk_index: The index of the requested chunk
        :type chunk_index: int
        :param chunk_size: The size of the requested chunk
        :type chunk_size: int
        """
        self.logger.debug("_on_packet_request started")
        self._wolk._on_packet_request(file_name, chunk_index, chunk_size)
        self.logger.debug("_on_packet_request ended")

    def _on_status(self, status):
        """
        Report firmware update status to WolkAbout IoT Platform.

        :param status: The status to be reported to the platform
        :type status: FirmwareStatus
        """
        self.logger.debug("_on_status started")
        self._wolk._on_status(status)
        self.logger.debug("_on_status ended")
